"""
The pii module includes utilities for working with Personally Identifying Information.

See the redactor.Redactor constructor for a list of PII recognizers being used.
"""

from .redactor import Redactor
